import datetime
import os
import shutil
from typing import Union

from metabolights_utils.utils.search_utils import MetabolightsSearchUtils as SearchUtils


class MetabolightsAuditUtils(object):

    @staticmethod
    def create_audit_folder(
        src_root_path: str,
        target_root_path: str,
        folder_suffix: Union[None, str] = "BACKUP",
        folder_prefix: Union[None, str] = None,
        timestamp_format: str = "%Y-%m-%d_%H-%M-%S",
    ) -> Union[None, str]:
        metadata_files_list = SearchUtils.get_isa_metadata_files(
            folder_path=src_root_path, recursive=False
        )
        metadata_files_list.sort()

        if metadata_files_list:
            base = datetime.datetime.now(datetime.timezone.utc).strftime(
                timestamp_format
            )
            folder_name = f"{base}_{folder_suffix}" if folder_suffix else base
            folder_name = (
                f"{folder_prefix}_{folder_name}" if folder_prefix else folder_name
            )

            target_folder_path = os.path.join(target_root_path, folder_name)

            os.makedirs(target_folder_path, exist_ok=True)

            for file in metadata_files_list:
                basename = os.path.basename(file)
                target_file = os.path.join(target_folder_path, basename)
                shutil.copy2(file, target_file, follow_symlinks=False)
            return target_folder_path
        return None

    @staticmethod
    def copy_isa_metadata_files(
        src_folder_path: str, target_folder_path: str
    ) -> Union[None, str]:
        metadata_files_list = SearchUtils.get_isa_metadata_files(
            folder_path=src_folder_path, recursive=False
        )
        metadata_files_list.sort()

        if metadata_files_list:
            os.makedirs(target_folder_path, exist_ok=True)
            for file in metadata_files_list:
                basename = os.path.basename(file)
                target_file = os.path.join(target_folder_path, basename)
                shutil.copy2(file, target_file, follow_symlinks=False)
            return target_folder_path
        return None



from typing import List, Union

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.isa.common import Comment, IsaAbstractModel, IsaTabConfig

module_name = __name__


class BaseSection(IsaAbstractModel):
    comments: List[Comment] = []


class OntologySourceReference(IsaAbstractModel):
    isatab_config: Annotated[IsaTabConfig, Field(exclude=True)] = IsaTabConfig(
        field_order=[
            "source_name",
            "source_file",
            "source_version",
            "source_description",
        ]
    )

    source_name: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Source Name",
                "source_description": True,
            },
        ),
    ] = ""

    source_file: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Source File"},
        ),
    ] = ""
    source_version: Annotated[
        Union[str, int],
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Source Version"},
        ),
    ] = ""
    source_description: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Source Description"},
        ),
    ] = ""


class OntologyAnnotation(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(exclude=True),
    ] = IsaTabConfig(field_order=["term", "term_accession_number", "term_source_ref"])

    term: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
            },
        ),
    ] = ""

    term_accession_number: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Term Accession Number",
            },
        ),
    ] = ""
    term_source_ref: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Term Source REF"},
        ),
    ] = ""


class ValueTypeAnnotation(IsaAbstractModel):
    isatab_config: Annotated[IsaTabConfig, Field(exclude=True)] = IsaTabConfig(
        field_order=[
            "name",
            "type",
            "term_accession_number",
            "term_source_ref",
        ]
    )

    name: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Name"}
        ),
    ] = ""
    type: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Type"}
        ),
    ] = ""
    term_accession_number: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Type Term Accession Number",
            },
        ),
    ] = ""
    term_source_ref: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Type Term Source REF",
            },
        ),
    ] = ""


class Publication(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["pub_med_id", "doi", "author_list", "title", "status"]
    )

    pub_med_id: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "PubMed ID"},
        ),
    ] = ""
    doi: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Publication DOI"},
        ),
    ] = ""
    author_list: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Publication Author List",
            },
        ),
    ] = ""
    title: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Publication Title"},
        ),
    ] = ""
    status: Annotated[
        OntologyAnnotation,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Publication Status"},
        ),
    ] = OntologyAnnotation()


class Person(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "last_name",
            "first_name",
            "mid_initials",
            "email",
            "phone",
            "fax",
            "address",
            "affiliation",
            "roles",
        ]
    )

    last_name: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Last Name"},
        ),
    ] = ""
    first_name: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "First Name"},
        ),
    ] = ""
    mid_initials: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Mid Initials"},
        ),
    ] = ""
    email: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Email"},
        ),
    ] = ""
    phone: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Phone"},
        ),
    ] = ""
    fax: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Fax"}
        ),
    ] = ""
    address: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Address"},
        ),
    ] = ""
    affiliation: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Affiliation"},
        ),
    ] = ""
    roles: Annotated[
        List[OntologyAnnotation],
        Field(
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Roles",
                "text_multiple_value": True,
                "seperator": ";",
            },
        ),
    ] = []


class Factor(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(field_order=["name", "type"])

    name: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Name"}
        ),
    ] = ""
    type: Annotated[
        OntologyAnnotation,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Type"},
        ),
    ] = OntologyAnnotation()


class Assay(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "file_name",
            "measurement_type",
            "technology_type",
            "technology_platform",
        ]
    )

    file_name: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "File Name"},
        ),
    ] = ""
    measurement_type: Annotated[
        OntologyAnnotation,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Measurement Type"},
        ),
    ] = OntologyAnnotation()
    technology_type: Annotated[
        OntologyAnnotation,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Technology Type"},
        ),
    ] = OntologyAnnotation()
    technology_platform: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Technology Platform"},
        ),
    ] = ""


class Protocol(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "name",
            "protocol_type",
            "description",
            "uri",
            "version",
            "parameters",
            "components",
        ]
    )

    name: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Name"}
        ),
    ] = ""
    protocol_type: Annotated[
        OntologyAnnotation,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Type"},
        ),
    ] = OntologyAnnotation()
    description: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Description"},
        ),
    ] = ""
    uri: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "URI"}
        ),
    ] = ""
    version: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Version"},
        ),
    ] = ""
    parameters: Annotated[
        List[OntologyAnnotation],
        Field(
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Parameters Name",
                "text_multiple_value": True,
                "seperator": ";",
            },
        ),
    ] = []
    components: Annotated[
        List[ValueTypeAnnotation],
        Field(
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Components",
                "text_multiple_value": True,
                "seperator": ";",
            },
        ),
    ] = []


######################################################################################################
# ISA TAB FILE SECTIONS
######################################################################################################


class OntologySourceReferences(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["references"],
        section_header="ONTOLOGY SOURCE REFERENCE",
        section_prefix="Term",
    )

    references: Annotated[
        List[OntologySourceReference],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Source Name",
            },
        ),
    ] = []


class InvestigationPublications(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["publications"],
        section_header="INVESTIGATION PUBLICATIONS",
        section_prefix=None,
    )

    publications: Annotated[
        List[Publication],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Publication Title",
            },
        ),
    ] = []


class InvestigationContacts(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["people"],
        section_header="INVESTIGATION CONTACTS",
        section_prefix="Person",
    )

    people: Annotated[
        List[Person],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Last Name",
            },
        ),
    ] = []


class StudyDesignDescriptors(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["design_types"],
        section_header="STUDY DESIGN DESCRIPTORS",
        section_prefix="Design",
    )

    design_types: Annotated[
        List[OntologyAnnotation],
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Type"}
        ),
    ] = []


class StudyPublications(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["publications"],
        section_header="STUDY PUBLICATIONS",
        section_prefix=None,
    )

    publications: Annotated[
        List[Publication],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Publication Title",
            },
        ),
    ] = []


class StudyFactors(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["factors"],
        section_header="STUDY FACTORS",
        section_prefix="Factor",
    )

    factors: Annotated[
        List[Factor],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Name",
            },
        ),
    ] = []


class StudyAssays(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["assays"],
        section_header="STUDY ASSAYS",
        section_prefix="Assay",
    )

    assays: Annotated[
        List[Assay],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "File Name",
            },
        ),
    ] = []


class StudyProtocols(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["protocols"],
        section_header="STUDY PROTOCOLS",
        section_prefix="Protocol",
    )

    protocols: Annotated[
        List[Protocol],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Name",
            },
        ),
    ] = []


class StudyContacts(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["people"],
        section_header="STUDY CONTACTS",
        section_prefix="Person",
    )

    people: Annotated[
        List[Person],
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Last Name",
            },
        ),
    ] = []


class Study(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "identifier",
            "title",
            "description",
            "submission_date",
            "public_release_date",
            "file_name",
        ],
        section_header="STUDY",
        section_prefix="Study",
    )

    identifier: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Identifier"},
        ),
    ] = ""
    title: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Title"},
        ),
    ] = ""
    description: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Description"},
        ),
    ] = ""
    submission_date: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Submission Date"},
        ),
    ] = ""
    public_release_date: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Public Release Date"},
        ),
    ] = ""
    file_name: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "File Name"},
        ),
    ] = ""

    study_design_descriptors: Annotated[
        StudyDesignDescriptors,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = StudyDesignDescriptors()
    study_publications: Annotated[
        StudyPublications,
        Field(description="", json_schema_extra={"auto_fill": True, "header_name": ""}),
    ] = StudyPublications()
    study_factors: Annotated[
        StudyFactors,
        Field(description="", json_schema_extra={"auto_fill": True, "header_name": ""}),
    ] = StudyFactors()
    study_assays: Annotated[
        StudyAssays,
        Field(description="", json_schema_extra={"auto_fill": True, "header_name": ""}),
    ] = StudyAssays()
    study_protocols: Annotated[
        StudyProtocols,
        Field(description="", json_schema_extra={"auto_fill": True, "header_name": ""}),
    ] = StudyProtocols()
    study_contacts: Annotated[
        StudyContacts,
        Field(description="", json_schema_extra={"auto_fill": True, "header_name": ""}),
    ] = StudyContacts()


class Investigation(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "identifier",
            "title",
            "description",
            "submission_date",
            "public_release_date",
        ],
        section_header="INVESTIGATION",
        section_prefix="Investigation",
        specification_version="1.0",
        specification_date="2016-10-28",
    )

    identifier: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Identifier"},
        ),
    ] = ""
    title: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Title"},
        ),
    ] = ""
    description: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Description"},
        ),
    ] = ""
    submission_date: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Submission Date"},
        ),
    ] = ""
    public_release_date: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={"auto_fill": True, "header_name": "Public Release Date"},
        ),
    ] = ""

    ontology_source_references: Annotated[
        OntologySourceReferences,
        Field(
            description="Ontology sources used in the investigation file",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "inherit_prefix": False,
            },
        ),
    ] = OntologySourceReferences()

    investigation_publications: Annotated[
        InvestigationPublications,
        Field(
            description="All publications prepared to report results of the investigation",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = InvestigationPublications()

    investigation_contacts: Annotated[
        InvestigationContacts,
        Field(
            description="People details of the investigation.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = InvestigationContacts()

    studies: Annotated[
        List[Study], Field(description="Studies carried out in the investigation.")
    ] = []



import glob
import os
from typing import List


class MetabolightsSearchUtils(object):
    @staticmethod
    def get_isa_metadata_files(folder_path: str, recursive=False) -> List[str]:
        metadata_files = []
        patterns = ["[asi]_*.txt", "m_*.tsv"]
        if not os.path.exists(folder_path) or not os.path.isdir(folder_path):
            return metadata_files

        for pattern in patterns:
            search_pattern = os.path.join(folder_path, pattern)
            metadata_files.extend(glob.glob(search_pattern, recursive=recursive))
        return metadata_files



from typing import Dict, List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.isa.common import AssayTechnique, IsaTableFile


class AssignmentFile(IsaTableFile):
    identified_metabolite_names: Annotated[List[str], Field(description="")] = []
    metabolite_assignments: Annotated[Dict[str, str], Field(description="")] = {}
    assay_technique: Annotated[AssayTechnique, Field(description="")] = AssayTechnique()
    number_of_rows: Annotated[int, Field(description="")] = 0
    number_of_assigned_rows: Annotated[int, Field(description="")] = 0
    number_of_unassigned_rows: Annotated[int, Field(description="")] = 0



import pathlib
import uuid
from typing import Dict, List

from metabolights_utils.tsv import model
from metabolights_utils.tsv.actions.base import BaseTsvAction


class AddRowsTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: model.TsvAddRowsAction,
    ) -> model.TsvActionResult:
        result: model.TsvActionResult = model.TsvActionResult(action=action)
        if action.action_type != model.TsvActionType.ADD_ROW:
            result.message = "Action name is not valid"
            return result

        action: model.TsvAddRowsAction = action
        target_row_indices: List[int] = action.new_row_indices

        if not target_row_indices:
            result.message = "There is not row index"
            return result

        row_data: Dict[int, model.TsvRowData] = (
            action.row_data if action.row_data else {}
        )

        row_indices = target_row_indices.copy()
        row_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with open(source_file_path, "r", encoding="utf-8") as source:
                header_line = source.readline()
                header_names = header_line.strip().split("\t")
                empty_row = [""] * len(header_names)

                with open(target_file_path, "w", encoding="utf-8") as target:
                    self.write_row(target, header_names)
                    row_index = -1
                    for line in source:
                        row_index += 1
                        if row_index in row_indices:
                            while row_index in row_indices:
                                input_row = (
                                    row_data[row_index]
                                    if row_index in row_data
                                    else None
                                )
                                if not input_row:
                                    self.write_row(target, empty_row)
                                else:
                                    new_row = self.get_updated_row(empty_row, input_row)
                                    self.write_row(target, new_row)
                                row_indices.remove(row_index)
                                row_index += 1
                        target.write(line)

                    if len(row_indices):
                        row_index += 1
                        while row_index in row_indices:
                            input_row = (
                                row_data[row_index] if row_index in row_data else None
                            )
                            if not input_row:
                                self.write_row(target, empty_row)
                            else:
                                new_row = self.get_updated_row(empty_row, input_row)
                                self.write_row(target, new_row)
                            row_indices.remove(row_index)
                            row_index += 1

                        if len(row_indices):
                            result.message = (
                                f"Invalid row indices {', '.join(row_indices)}"
                            )
                            return result
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class CamelCaseModel(BaseModel):
    """Base model class to convert python attributes to camel case"""

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        json_schema_serialization_defaults_required=True,
    )



from typing import List

from metabolights_utils.isatab.default.base_isa_table_file import BaseIsaTableFileReader


class DefaultAssayFileReader(BaseIsaTableFileReader):
    patterns = [
        [r"^(Extract Name)$", ""],
        [r"^(Protocol REF)(.\d+)?$", "Protocol"],
        [r"^(Sample Name)$", ""],
        [r"^[ ]*Parameter[ ]+Value[ ]*\[[ ]*(\w[ -~]*)[ ]*\][ ]*$", "Parameter Value"],
        [r"^[ ]*Comment[ ]*\[[ ]*(\w[ -~]*)[ ]*\][ ]*$", "Comment"],
        [r"^(Labeled Extract Name)$", ""],
        [r"^(Label)$", ""],
    ]

    def __init__(self, results_per_page=100) -> None:
        super().__init__(results_per_page=results_per_page)

    def get_expected_patterns(self) -> List[List[str]]:
        return DefaultAssayFileReader.patterns



import pathlib
from abc import ABC, abstractmethod
from io import IOBase
from typing import List, Union

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.common import MetabolightsBaseModel
from metabolights_utils.models.isa.common import IsaTableFile
from metabolights_utils.models.isa.investigation_file import Investigation
from metabolights_utils.models.parser.common import ParserReport
from metabolights_utils.tsv.filter import TsvFileFilterOption
from metabolights_utils.tsv.sort import TsvFileSortOption


class IsaTableFileReaderResult(MetabolightsBaseModel):
    isa_table_file: Annotated[IsaTableFile, Field()] = IsaTableFile()
    parser_report: Annotated[ParserReport, Field()] = ParserReport()


class InvestigationFileReaderResult(MetabolightsBaseModel):
    investigation: Annotated[Investigation, Field()] = Investigation()
    parser_report: Annotated[ParserReport, Field()] = ParserReport()
    file_path: str = ""
    sha256_hash: Union[None, str] = None


class InvestigationFileReader(ABC):
    @abstractmethod
    def read(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        skip_parser_info_messages: bool = True,
    ) -> InvestigationFileReaderResult:
        """Reads investigation an file and return Investigation model and parser messages.
            If file buffer is not defined, it reads file_path.
           At least one of the file_buffer and file_path parameters should be defined.

        Args:
            file_buffer_or_path (str, pathlib.Path, IOBase): File buffer to read file content. io.StringIO, io.TextIOWrapper with open(), etc. Defaults to None.
            file_path (Union[str, pathlib.Path], optional): File path or pathlib.Path object. Defaults to None.
            skip_parser_info_messages (bool, optional): clear INFO messages from parser messages. Defaults to True.

        Raises:
            exc: any unexpected exception while reading file

        Returns:
            InvestigationFileReaderResult: Investigation model and parser messages
        """


class IsaTableFileReader(ABC):
    @abstractmethod
    def get_headers(
        self, file_buffer_or_path: Union[str, pathlib.Path, IOBase]
    ) -> IsaTableFileReaderResult:
        """_summary_

        Args:
            file_buffer (Union[str, pathlib.Path, IOBase]): File buffer to read file content. io.StringIO, io.TextIOWrapper with open(), etc.
            file_path (Union[str, pathlib.Path], optional): File path or pathlib.Path object.

        Returns:
            IsaTableFileWrapperResult: Returns IsaTableFile without rows and parser messages.
        """

    @abstractmethod
    def get_total_pages(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        results_per_page: int = 100,
    ) -> int:
        """Returns total number of pages according to results_per_page. At least one of the file_buffer and file_path parameters should be defined.
           If file_buffer is not defined, file_path is not used to read content.
           If file_path is not defined, file path and name will not be added to messages.

        Args:
            file_buffer (Union[str, pathlib.Path, IOBase]): File buffer to read file content. io.StringIO, io.TextIOWrapper with open(), etc.
            results_per_page (int): Items per page. Defaults: 100.

        Returns:
            int: total number of pages in file.
        """

    @abstractmethod
    def get_total_row_count(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
    ) -> int:
        """Returns total number of rows. At least one of the file_buffer and file_path parameters should be defined.
           If file_buffer is not defined, file_path is not used to read content.
           If file_path is not defined, file path and name will not be added to messages.

        Args:
            file_buffer_or_path (str, pathlib.Path, IOBase): File buffer to read file content. io.StringIO, io.TextIOWrapper with open(), etc.
            file_path (Union[str, pathlib.Path], optional): File path or pathlib.Path object.

        Returns:
            int: total number of rows in file. First column is assigned as header row.
        """

    @abstractmethod
    def get_page(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        page: int = 1,
        results_per_page: int = 100,
        selected_columns: Union[List[str], None] = None,
        filter_options: List[TsvFileFilterOption] = None,
        sort_options: List[TsvFileSortOption] = None,
    ) -> IsaTableFileReaderResult:
        """Reads file and returns the selected page of file. At least one of the file_buffer and file_path parameters should be defined.
           If file_buffer is not defined, file_path is not used to read content.
           If file_path is not defined, file path and name will not be added to messages.

        Args:
            file_buffer_or_path (str, pathlib.Path, IOBase): File buffer to read file content. io.StringIO, io.TextIOWrapper with open(), etc.
            page (int, optional): The page number requested. Defaults to 1.
            results_per_page (int, optional): Number of rows in each page. Defaults to 100.
            selected_columns (Union[List[str], None], optional): Column names will be returned. Returns all columns if it is None. Defaults to None.
            file_path (Union[str, pathlib.Path], optional): File path str or pathlib.Path object
            filter_options (List[TsvFileFilterOption]): filter column names and filter methods. Defaults to None.
            sort_options (List[TsvFileSortOption]): Sort column names. Defaults to None.
        Returns:
            IsaTableFileReaderResult: IsaTableFile model and parser messages
        """

    @abstractmethod
    def get_rows(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        offset: int = 0,
        limit: Union[int, None] = None,
        selected_columns: Union[None, List[str]] = None,
        filter_options: List[TsvFileFilterOption] = None,
        sort_options: List[TsvFileSortOption] = None,
    ) -> IsaTableFileReaderResult:
        """Reads file and returns rows of the file starting from offset.
           At least one of the file_buffer and file_path parameters should be defined.
           If limit is defined, size of the returned rows will be limited to this value.

        Args:
            file_buffer_or_path (str, pathlib.Path, IOBase): File buffer to read file content. io.StringIO, io.TextIOWrapper with open(), etc.
            offset (int, optional): Starting index of rows will be returned. First rows is header and index of second row is 0. Defaults to 0.
            limit (Union[int, None], optional): Number of rows will be returned. If it is None, return all rows. Defaults to None.
            column_names (Union[List[str], None], optional): Column names will be returned. Returns all columns if it is None. Defaults to None.
            filter_options (List[TsvFileFilterOption]): filter column names and filter methods. Defaults to None.
            sort_options (List[TsvFileSortOption]): Sort column names. Defaults to None.

        Returns:
            IsaTableFileReaderResult: IsaTableFile model and parser messages.
        """

    @abstractmethod
    def read(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        offset: Union[None, int],
        limit: Union[None, int],
        selected_columns: Union[None, List[str]] = None,
        skip_parser_info_messages: bool = True,
        filter_options: List[TsvFileFilterOption] = None,
        sort_options: List[TsvFileSortOption] = None,
    ) -> IsaTableFileReaderResult:
        """Reads selected rows with selected columns from tsv file. If sort and filter options are enabled, offset and limit are applied to the final filter and sort result.

        Args:
            file_buffer_or_path (Union[str, pathlib.Path, IOBase]): File buffer or path to read file content. file path str, io.StringIO, io.TextIOWrapper with open(), etc.
            offset (int, optional): Starting index of rows will be returned. First rows is header and index of the second row is 0. Defaults to 0.
            limit (Union[int, None], optional): Number of rows will be returned. If it is None, return all rows. Defaults to 1000.
            selected_columns (Union[List[str], None], optional): Column names will be returned. Returns all columns if it is None. Defaults to None.
            skip_parser_info_messages (bool, optional): clear INFO messages from parser messages. Defaults to True.
            filter_options (List[TsvFileFilterOption]): filter column names and filter methods. Defaults to None.
            sort_options (List[TsvFileSortOption]): Sort column names. Defaults to None.
        Returns:
            IsaTableFileReaderResult: IsaTableFile model and parser messages.
        """


class IsaTabReaderFactory(ABC):
    @abstractmethod
    def get_investigation_file_reader(self) -> InvestigationFileReader:
        pass

    @abstractmethod
    def get_assay_file_reader(self, results_per_page=100) -> IsaTableFileReader:
        pass

    @abstractmethod
    def get_sample_file_reader(self, results_per_page=100) -> IsaTableFileReader:
        pass

    @abstractmethod
    def get_assignment_file_reader(self, results_per_page=100) -> IsaTableFileReader:
        pass



from metabolights_utils.common import CamelCaseModel
from metabolights_utils.isatab import Reader, Writer
from metabolights_utils.isatab.default.assay_file import DefaultAssayFileReader
from metabolights_utils.isatab.default.assignment_file import (
    DefaultAssignmentFileReader,
)
from metabolights_utils.isatab.default.factory import (
    DefaultIsaTabReaderFactory,
    DefaultIsaTabWriterFactory,
)
from metabolights_utils.isatab.default.investigation_file import (
    DefaultInvestigationFileReader,
    DefaultInvestigationFileWriter,
)
from metabolights_utils.isatab.default.parser.isa_table_parser import (
    assay_file_expected_patterns,
    samples_file_expected_patterns,
)
from metabolights_utils.isatab.default.sample_file import DefaultSampleFileReader
from metabolights_utils.isatab.default.writer import DefaultIsaTableFileWriter
from metabolights_utils.isatab.reader import (
    InvestigationFileReader,
    InvestigationFileReaderResult,
    IsaTableFileReader,
    IsaTableFileReaderResult,
    IsaTabReaderFactory,
)
from metabolights_utils.isatab.writer import (
    InvestigationFileWriter,
    IsaTableFileWriter,
    IsaTabWriterFactory,
)
from metabolights_utils.models.common import GenericMessage, MetabolightsBaseModel
from metabolights_utils.models.enums import GenericMessageType
from metabolights_utils.models.isa.assay_file import AssayFile
from metabolights_utils.models.isa.assignment_file import AssignmentFile
from metabolights_utils.models.isa.common import (
    AssayTechnique,
    Comment,
    IsaAbstractModel,
    IsaTable,
    IsaTableColumn,
    IsaTableFile,
    NumericItem,
    OntologyItem,
    OrganismAndOrganismPartPair,
)
from metabolights_utils.models.isa.enums import (
    ColumnsStructure,
    IsaTableAdditionalColumn,
)
from metabolights_utils.models.isa.investigation_file import (
    Assay,
    BaseSection,
    Factor,
    Investigation,
    InvestigationContacts,
    InvestigationPublications,
    OntologyAnnotation,
    OntologySourceReference,
    OntologySourceReferences,
    Person,
    Protocol,
    Publication,
    Study,
    StudyAssays,
    StudyContacts,
    StudyFactors,
    StudyProtocols,
    StudyPublications,
    ValueTypeAnnotation,
)
from metabolights_utils.models.metabolights.metabolights_study import (
    BaseMetabolightsStudyModel,
)
from metabolights_utils.models.parser.common import ParserMessage, ParserReport
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.tsv.actions.add_column import AddColumnsTsvAction
from metabolights_utils.tsv.actions.add_row import AddRowsTsvAction
from metabolights_utils.tsv.actions.base import BaseTsvAction, TsvActionException
from metabolights_utils.tsv.actions.copy_column import CopyColumnTsvAction
from metabolights_utils.tsv.actions.copy_row import CopyRowTsvAction
from metabolights_utils.tsv.actions.delete_column import DeleteColumnsTsvAction
from metabolights_utils.tsv.actions.delete_row import DeleteRowsTsvAction
from metabolights_utils.tsv.actions.move_column import MoveColumnTsvAction
from metabolights_utils.tsv.actions.update_cell import UpdateCellsTsvAction
from metabolights_utils.tsv.actions.update_column import UpdateColumnsTsvAction
from metabolights_utils.tsv.actions.update_column_header import (
    UpdateColumnHeadersTsvAction,
)
from metabolights_utils.tsv.actions.update_row import UpdateRowsTsvAction
from metabolights_utils.tsv.filter import (
    BetweenEqualCustomFilter,
    ContainsFilter,
    CustomFilter,
    EmptyFilter,
    EndsWithFilter,
    EnumContainsCustomFilter,
    Filter,
    FilterDataType,
    FilterOperation,
    FilterRegistry,
    GreaterEqualFilter,
    GreaterFilter,
    LessEqualFilter,
    LessFilter,
    RegexFilter,
    StartsWithFilter,
    TsvFileFilterOption,
    TsvFilterException,
    ValidDatetimeCustomFilter,
    ValidNumberCustomFilter,
)
from metabolights_utils.tsv.model import (
    TsvAction,
    TsvActionReport,
    TsvActionResult,
    TsvActionType,
    TsvAddColumnsAction,
    TsvAddRowsAction,
    TsvCellData,
    TsvColumnData,
    TsvCopyColumnAction,
    TsvCopyRowAction,
    TsvDeleteColumnsAction,
    TsvDeleteRowsAction,
    TsvMoveColumnAction,
    TsvMoveRowAction,
    TsvRowData,
    TsvUpdateCellsAction,
    TsvUpdateColumnHeaderAction,
    TsvUpdateColumnsAction,
    TsvUpdateRowsAction,
)
from metabolights_utils.tsv.sort import (
    AbstractSorter,
    CustomSorter,
    DateTimeSorter,
    EnumSorter,
    FloatSorter,
    IntegerSorter,
    Sorter,
    SorterRegistry,
    SortType,
    SortValueClassification,
    StringSorter,
    TsvFileSortOption,
    TsvFileSortValueOrder,
    TsvSortException,
)
from metabolights_utils.tsv.tsv_file_updater import TSV_FILE_ACTIONS, TsvFileUpdater
from metabolights_utils.utils.audit_utils import MetabolightsAuditUtils
from metabolights_utils.utils.filename_utils import MetabolightsFileNamehUtils
from metabolights_utils.utils.hash_utils import (
    EMPTY_FILE_HASH,
    IsaMetadataFolderHash,
    MetabolightsHashUtils,
)
from metabolights_utils.utils.search_utils import MetabolightsSearchUtils

__all__ = [
    "Assay",
    "BaseSection",
    "Factor",
    "Investigation",
    "InvestigationContacts",
    "InvestigationPublications",
    "OntologyAnnotation",
    "OntologySourceReference",
    "OntologySourceReferences",
    "Person",
    "Protocol",
    "Publication",
    "Study",
    "StudyAssays",
    "StudyContacts",
    "StudyFactors",
    "StudyProtocols",
    "StudyPublications",
    "ValueTypeAnnotation",
    "ColumnsStructure",
    "IsaTableAdditionalColumn",
    "IsaAbstractModel",
    "NumericItem",
    "AssayFile",
    "AssignmentFile",
    "GenericMessage",
    "GenericMessageType",
    "MetabolightsBaseModel",
    "CamelCaseModel",
    "AssayTechnique",
    "Comment",
    "IsaTable",
    "IsaTableColumn",
    "IsaTableFile",
    "OntologyItem",
    "OrganismAndOrganismPartPair",
    "BaseMetabolightsStudyModel",
    "ParserMessage",
    "ParserMessageType",
    "ParserReport",
    "Reader",
    "DefaultAssayFileReader",
    "InvestigationFileReader",
    "DefaultAssignmentFileReader",
    "IsaTableFileReader",
    "InvestigationFileReaderResult",
    "IsaTableFileReaderResult",
    "DefaultIsaTabReaderFactory",
    "DefaultIsaTabWriterFactory",
    "DefaultInvestigationFileReader",
    "DefaultInvestigationFileWriter",
    "assay_file_expected_patterns",
    "samples_file_expected_patterns",
    "DefaultSampleFileReader",
    "DefaultIsaTableFileWriter",
    "IsaTabReaderFactory",
    "Writer",
    "InvestigationFileWriter",
    "IsaTableFileWriter",
    "IsaTabWriterFactory",
    # TSV
    "TsvAction",
    "TsvActionReport",
    "TsvActionResult",
    "TsvActionType",
    "TsvAddColumnsAction",
    "TsvAddRowsAction",
    "TsvCellData",
    "TsvColumnData",
    "TsvCopyColumnAction",
    "TsvCopyRowAction",
    "TsvDeleteColumnsAction",
    "TsvDeleteRowsAction",
    "TsvMoveColumnAction",
    "TsvMoveRowAction",
    "TsvRowData",
    "TsvUpdateCellsAction",
    "TsvUpdateColumnHeaderAction",
    "TsvUpdateColumnsAction",
    "TsvUpdateRowsAction",
    "Filter",
    "FilterDataType",
    "FilterOperation",
    "TsvFileFilterOption",
    "TsvFilterException",
    "CustomFilter",
    "FilterRegistry",
    "ContainsFilter",
    "StartsWithFilter",
    "EndsWithFilter",
    "GreaterFilter",
    "GreaterEqualFilter",
    "LessFilter",
    "LessEqualFilter",
    "EmptyFilter",
    "RegexFilter",
    "ValidNumberCustomFilter",
    "ValidDatetimeCustomFilter",
    "EnumContainsCustomFilter",
    "BetweenEqualCustomFilter",
    "AbstractSorter",
    "CustomSorter",
    "DateTimeSorter",
    "EnumSorter",
    "FloatSorter",
    "IntegerSorter",
    "Sorter",
    "SorterRegistry",
    "SortType",
    "SortValueClassification",
    "StringSorter",
    "TsvFileSortOption",
    "TsvFileSortValueOrder",
    "TsvSortException",
    "TSV_FILE_ACTIONS",
    "TsvFileUpdater",
    "TsvActionException",
    "BaseTsvAction",
    "AddColumnsTsvAction",
    "AddRowsTsvAction",
    "CopyColumnTsvAction",
    "CopyRowTsvAction",
    "DeleteColumnsTsvAction",
    "DeleteRowsTsvAction",
    "MoveColumnTsvAction",
    "UpdateCellsTsvAction",
    "UpdateColumnsTsvAction",
    "UpdateColumnHeadersTsvAction",
    "UpdateRowsTsvAction",
    # utils
    "MetabolightsAuditUtils",
    "MetabolightsSearchUtils",
    "MetabolightsFileNamehUtils",
    # hash utils
    "MetabolightsHashUtils",
    "IsaMetadataFolderHash",
    "EMPTY_FILE_HASH",
]



import pathlib
from typing import Dict, Union

from metabolights_utils.isatab.writer import IsaTableFileWriter
from metabolights_utils.models.isa.common import IsaTable
from metabolights_utils.tsv.model import (
    TsvActionReport,
    TsvColumnData,
    TsvUpdateColumnsAction,
)
from metabolights_utils.tsv.tsv_file_updater import TsvFileUpdater


class DefaultIsaTableFileWriter(TsvFileUpdater, IsaTableFileWriter):
    def save_isa_table(
        self,
        file_path: Union[str, pathlib.Path],
        file_sha256_hash: str,
        isa_table: IsaTable,
    ) -> TsvActionReport:
        column_indices = dict(zip(isa_table.columns, isa_table.column_indices))
        headers = {
            column_indices[x.column_name]: x.column_header for x in isa_table.headers
        }

        columns: Dict[int, TsvColumnData] = {}
        for column_name in isa_table.data:
            column_data: TsvColumnData = TsvColumnData()
            column_index = column_indices[column_name]
            columns[column_index] = column_data
            column_data.header_name = headers[column_index]
            column_data.values = dict(
                zip(isa_table.row_indices, isa_table.data[column_name])
            )

        column_update_action = TsvUpdateColumnsAction(columns=columns)
        report = self.apply_actions(
            file_path=file_path,
            file_sha256_hash=file_sha256_hash,
            actions=[column_update_action],
        )
        return report



import os
import pathlib
import sys
from io import IOBase
from typing import List, Union

from pydantic import BaseModel
from pydantic.alias_generators import to_camel, to_snake

from metabolights_utils.isatab.default.base_isa_file import BaseIsaFile
from metabolights_utils.isatab.default.parser.investigation_parser import (
    get_investigation,
)
from metabolights_utils.isatab.reader import (
    InvestigationFileReader,
    InvestigationFileReaderResult,
)
from metabolights_utils.isatab.writer import InvestigationFileWriter
from metabolights_utils.models.isa.common import IsaAbstractModel
from metabolights_utils.models.isa.investigation_file import BaseSection, Investigation
from metabolights_utils.models.isa.investigation_file import (
    module_name as inv_module_name,
)
from metabolights_utils.models.parser.common import ParserMessage, ParserReport
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.utils.hash_utils import MetabolightsHashUtils as HashUtils


class InvestigationFileException(Exception):
    def __init__(self, message) -> None:
        self.message = message


class DefaultInvestigationFileReader(InvestigationFileReader, BaseIsaFile):
    def read(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        skip_parser_info_messages: bool = True,
    ) -> InvestigationFileReaderResult:
        buffer_or_path, path = self._get_file_path(file_buffer_or_path)
        investigation = Investigation()
        parse_success = False
        read_messages: List[ParserMessage] = []
        try:
            file_buffer = self._get_file_buffer(buffer_or_path)
            if isinstance(file_buffer, IOBase):
                investigation = get_investigation(
                    file_buffer, path, messages=read_messages
                )
            else:
                investigation = get_investigation(None, path, messages=read_messages)
            messages = read_messages
            parse_success = True
        except UnicodeDecodeError as err:
            try:
                file_buffer = self._get_file_buffer(buffer_or_path, encoding="ascii")
                if isinstance(file_buffer, IOBase):
                    investigation = get_investigation(
                        file_buffer, path, messages=read_messages
                    )
                else:
                    investigation = get_investigation(
                        None, path, messages=read_messages
                    )
                messages = read_messages
                parse_success = True
            except Exception as exc:
                read_messages.append(
                    ParserMessage(
                        type=ParserMessageType.CRITICAL,
                        short="Parse error",
                        detail=(str(exc)),
                    )
                )
        except Exception as exc:
            read_messages.append(
                ParserMessage(
                    type=ParserMessageType.CRITICAL,
                    short="Parse error",
                    detail=(str(exc)),
                )
            )
        finally:
            self._close_file(file_buffer_or_path)

        if skip_parser_info_messages:
            messages = [x for x in read_messages if x.type != ParserMessageType.INFO]
        report = ParserReport(messages=messages)

        result = InvestigationFileReaderResult(
            investigation=investigation, parser_report=report
        )

        if parse_success:
            if pathlib.Path(path).exists():
                result.sha256_hash = HashUtils.sha256sum(path)
            elif (
                isinstance(buffer_or_path, str)
                or isinstance(buffer_or_path, pathlib.Path)
            ) and pathlib.Path(str(buffer_or_path)).exists():
                result.sha256_hash = HashUtils.sha256sum(str(buffer_or_path))
        return result


class DefaultInvestigationFileWriter(InvestigationFileWriter, BaseIsaFile):
    def write(
        self,
        investigation: Investigation,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase] = None,
        values_in_quotation_mark: bool = True,
        verify_file_after_update: bool = True,
        skip_parser_info_messages: bool = True,
    ) -> InvestigationFileReaderResult:
        content = InvestigationFileSerializer.to_isa_file_string(
            investigation=investigation,
            values_in_quotation_mark=values_in_quotation_mark,
        )
        file_path = None
        try:
            if isinstance(file_buffer_or_path, IOBase):
                file_buffer_or_path.write(content)
                file_path = file_buffer_or_path.name
            else:
                file_path = str(file_buffer_or_path)
                with open(file_buffer_or_path, "w", encoding="utf-8") as f:
                    f.write(content)
            report = ParserReport()
            if verify_file_after_update:
                read_messages: List[ParserMessage] = []
                if isinstance(file_buffer_or_path, IOBase):
                    investigation = get_investigation(
                        file_buffer_or_path,
                        file_buffer_or_path.name,
                        messages=read_messages,
                    )
                else:
                    investigation = get_investigation(
                        None, file_buffer_or_path, messages=read_messages
                    )
                messages = read_messages
                if skip_parser_info_messages:
                    messages = [
                        x for x in read_messages if x.type != ParserMessageType.INFO
                    ]
                report = ParserReport(messages=messages)

                result = InvestigationFileReaderResult(
                    investigation=investigation,
                    parser_report=report,
                    file_path=str(file_path),
                )
            if os.path.exists(file_path):
                result.sha256_hash = HashUtils.sha256sum(file_path)

            return result
        except Exception as exc:
            raise exc


class InvestigationFileSerializer(object):
    @classmethod
    def to_isa_file_lines(
        cls, investigation: Investigation, values_in_quotation_mark: bool = False
    ):
        file_lines = []
        for row in cls.to_isa_file_line_list(investigation):
            for i in range(len(row)):
                item = row[i].strip('"')
                row[i] = f'"{item}"' if i > 0 and values_in_quotation_mark else item
            file_lines.append("\t".join(row))

        return file_lines

    @classmethod
    def to_isa_file_string(
        cls, investigation: Investigation, values_in_quotation_mark: bool = False
    ):
        file_lines = cls.to_isa_file_lines(investigation, values_in_quotation_mark)

        return "\n".join(file_lines) + "\n"

    @classmethod
    def to_isa_file_line_list(cls, investigation: Investigation):
        rows: List[List[str]] = []
        rows.extend(cls.add_sub_section("", investigation.ontology_source_references))
        rows.extend(cls.add_sub_section("", investigation))
        rows.extend(
            cls.add_sub_section(
                investigation.isatab_config.section_prefix,
                investigation.investigation_publications,
            )
        )
        rows.extend(
            cls.add_sub_section(
                investigation.isatab_config.section_prefix,
                investigation.investigation_contacts,
            )
        )
        for study in investigation.studies:
            rows.extend(cls.add_sub_section("", study))
            rows.extend(
                cls.add_sub_section(
                    study.isatab_config.section_prefix, study.study_design_descriptors
                )
            )
            rows.extend(
                cls.add_sub_section(
                    study.isatab_config.section_prefix, study.study_publications
                )
            )
            rows.extend(
                cls.add_sub_section(
                    study.isatab_config.section_prefix, study.study_factors
                )
            )
            rows.extend(
                cls.add_sub_section(
                    study.isatab_config.section_prefix, study.study_assays
                )
            )
            rows.extend(
                cls.add_sub_section(
                    study.isatab_config.section_prefix, study.study_protocols
                )
            )
            rows.extend(
                cls.add_sub_section(
                    study.isatab_config.section_prefix, study.study_contacts
                )
            )
        return rows

    @staticmethod
    def get_attribute(model, field_name):
        model_field_name = to_snake(field_name)
        return getattr(model, model_field_name)

    @classmethod
    def add_sub_section(cls, prefix, model: BaseSection):
        rows: List[List[str]] = [[model.isatab_config.section_header]]
        header = []
        if prefix:
            header.append(prefix)
        data_schema = model.model_json_schema()
        field_order = model.isatab_config.field_order
        if not field_order:
            field_order = []
        header_name = model.isatab_config.section_prefix
        if header_name:
            header.append(header_name)

        for field_name in field_order:
            value = cls.get_attribute(model, field_name)
            field_key = to_camel(field_name)
            if isinstance(value, list):
                properties = data_schema["properties"]
                if "allOf" in properties[field_key] or "items" in properties[field_key]:
                    if "allOf" in properties[field_key]:
                        item_ref = properties[field_key]["allOf"][0]["$ref"]
                    else:
                        item_ref = properties[field_key]["items"]["$ref"]
                    default_object_name = item_ref.replace(
                        "#/definitions/", ""
                    ).replace("#/$defs/", "")
                    obj: BaseModel = getattr(
                        sys.modules[inv_module_name], default_object_name
                    )
                    props = obj.model_json_schema()["properties"]
                    header.append(data_schema["properties"][field_key]["header_name"])
                    sub_model_rows = cls.add_model_content(
                        " ".join(header).strip(), value, props
                    )
                    if sub_model_rows:
                        rows.extend(sub_model_rows)
            elif isinstance(value, str):
                field_attribute = to_snake(field_name)
                header_name = model.model_fields[field_attribute].json_schema_extra[
                    "header_name"
                ]
                header_name = (
                    f"{model.isatab_config.section_prefix} {header_name}"
                    if model.isatab_config.section_prefix
                    else header_name
                )
                rows.append([header_name, value])
            else:
                raise InvestigationFileException(
                    message=f"Unsopported type {type(value)}. Value {str(value)}"
                )

        cls.add_comments(model, rows)
        return rows

    @classmethod
    def add_comments(cls, model: BaseSection, rows: List[List[str]]):
        for comment in model.comments:
            comment_line = [f"Comment[{comment.name}]"]
            rows.append(comment_line)
            for value in comment.value:
                comment_line.append(value)

    @classmethod
    def add_model_content(cls, prefix, items: List[IsaAbstractModel], properties):
        rows = []
        row_map = {}
        fields = properties["isatabConfig"]["default"]["fieldOrder"]
        for i in range(len(fields)):
            field = fields[i]
            field_key = to_camel(field)
            header_name = properties[field_key]["header_name"]
            header_name = f"{prefix} {header_name}".strip() if prefix else header_name
            if "allOf" in properties[field_key] or "items" in properties[field_key]:
                if "allOf" in "allOf" in properties[field_key]:
                    item_ref: str = properties[field_key]["allOf"][0]["$ref"]
                else:
                    item_ref: str = properties[field_key]["items"]["$ref"]
                class_name = item_ref.replace("#/definitions/", "").replace(
                    "#/$defs/", ""
                )
                obj: BaseModel = getattr(sys.modules[inv_module_name], class_name)
                props = obj.model_json_schema()["properties"]
                input_items = (
                    [cls.get_attribute(item, field_key) for item in items]
                    if items
                    else [obj()]
                )

                item_values = cls.add_model_content(header_name, input_items, props)
                rows.extend(item_values)
            elif "type" in properties[field_key]:
                cls.assign_type(
                    items, properties, rows, row_map, fields, i, header_name
                )
            elif "anyOf" in properties[field_key]:

                cls.assign_string_type(items, rows, row_map, fields, i, header_name)

        return rows

    @classmethod
    def assign_type(cls, items, properties, rows, row_map, fields, i, header_name):
        field_key = to_camel(fields[i])
        object_type = properties[field_key]["type"]
        if object_type == "string":
            row_map[header_name] = [header_name]
            rows.append(row_map[header_name])
            if items:
                for item in items:
                    if isinstance(item, list):
                        sub_items = []
                        if item:
                            for sub_item in item:
                                sub_items.append(cls.get_attribute(sub_item, field_key))
                        else:
                            sub_items.append("")
                        row_map[header_name].append(";".join(sub_items))
                    else:
                        value = cls.get_attribute(item, field_key)
                        row_map[header_name].append(value)
            else:
                row_map[header_name].append("")
        else:
            raise InvestigationFileException(
                message=f"Invalid object type {object_type}"
            )

    @classmethod
    def assign_string_type(cls, items, rows, row_map, fields, i, header_name):
        field_key = to_camel(fields[i])
        row_map[header_name] = [header_name]
        rows.append(row_map[header_name])
        if items:
            for item in items:
                if isinstance(item, list):
                    sub_items = []
                    if item:
                        for sub_item in item:
                            sub_items.append(
                                str(cls.get_attribute(sub_item, field_key))
                            )
                    else:
                        sub_items.append("")
                    row_map[header_name].append(";".join(sub_items))
                else:
                    value = str(cls.get_attribute(item, field_key))
                    row_map[header_name].append(value)
        else:
            row_map[header_name].append("")



from typing import List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.isa.common import (
    IsaTableFile,
    OntologyItem,
    OrganismAndOrganismPartPair,
)


class SamplesFile(IsaTableFile):
    sample_names: Annotated[List[str], Field(description="")] = []
    organisms: Annotated[List[OntologyItem], Field(description="")] = []
    organism_parts: Annotated[List[OntologyItem], Field(description="")] = []
    organism_and_organism_part_pairs: Annotated[
        List[OrganismAndOrganismPartPair], Field(description="")
    ] = []
    variants: Annotated[List[OntologyItem], Field(description="")] = []
    sample_types: Annotated[List[OntologyItem], Field(description="")] = []
